/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2010 France Telecom R&D
 * Contact: jonas-team@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id: MDBTransportAsyncHandlerInvoker.java 20716 2011-01-11 11:56:58Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.ow2.jonas.cxf.transport.mdb;

import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.namespace.QName;

import org.apache.cxf.common.logging.LogUtils;
import org.apache.cxf.message.Exchange;
import org.apache.cxf.message.Message;
import org.apache.cxf.service.invoker.Invoker;
import org.apache.cxf.service.model.BindingOperationInfo;
import org.ow2.jonas.cxf.transport.mdb.api.AsyncHandler;
import org.ow2.jonas.cxf.transport.mdb.api.ContextInfo;

/**
 * Invoker that will invoke all async handlers.
 * @author Florent Benoit
 */
public class MDBTransportAsyncHandlerInvoker implements Invoker {

    /**
     * Logger.
     */
    private static final Logger LOG = LogUtils.getL7dLogger(MDBTransportAsyncHandlerInvoker.class);

    
    /**
     * Invoke the registered async handlers with the result of an asynchronous call.
     * @param exchange the CXF exchange object
     * @param o the result of the call
     */
    public Object invoke(Exchange exchange, Object o) {
        BindingOperationInfo bop = exchange.get(BindingOperationInfo.class);

        // Get correlation ID
        Message message = exchange.getInMessage();
        String correlationID = (String) message.get("CorrelationID");
        if (correlationID == null) {
            throw new IllegalStateException("No correlationID has been found for operation '" + bop.getName() + "'.");
        }

        if (!(o instanceof List<?>)) {
            throw new IllegalArgumentException("Invoker parameter should be a list, found " + o + "' for operation '"
                    + bop.getName() + "'.");
        }
        List<?> lst = (List<?>) o;
        Object response = null;
        // Response is here
        if (lst.size() == 1) {
            response = lst.get(0);
        } else if (lst.size() == 0) {
            // void answer
            response = Void.TYPE;
        } else {
            throw new IllegalArgumentException("Invoker parameter should be a list with none or one parameter, found " + lst
                + "' for operation '" + bop.getName() + "'.");
        }

        // Retrieve AsyncHandlers from exchange object and invoke them
        MDBTransportConfiguration mdbTransportConfiguration = exchange.get(MDBTransportConfiguration.class);
        if (mdbTransportConfiguration == null) {
            throw new IllegalStateException("Unable to get the MDB receiver configuration for operation '" + bop.getName() + "'.");
        }
        List<AsyncHandler> asyncHandlers = mdbTransportConfiguration.getAsyncHandlers();
        if (asyncHandlers != null) {
            boolean distributed = false;
            for (AsyncHandler asyncHandler : asyncHandlers) {
                ContextInfo contextInfo = new ContextInfo();

                contextInfo.setCorrelationID(correlationID);
                contextInfo.setWsdlOperation(bop.getName());
                contextInfo.setWsdlService((QName) exchange.get("javax.xml.ws.wsdl.service"));
                contextInfo.setWsdlInterface((QName) exchange.get("javax.xml.ws.wsdl.interface"));
                contextInfo.setWsdlPort((QName) exchange.get("javax.xml.ws.wsdl.port"));

                boolean known = false;
                try {
                    known = asyncHandler.doYouKnowThisContext(contextInfo);
                } catch (Error e) {
                    LOG.log(Level.SEVERE, "Error while asking the asynchandler. continue", e);
                } catch (Exception e) {
                    LOG.log(Level.SEVERE, "Error while asking the asynchandler. continue", e);
                }
                
                if (known) {
                    asyncHandler.doHandlerResponse(response, contextInfo);
                    distributed = true;
                }
            }
            if (!distributed) {
                throw new IllegalStateException("Message has not been processed as no defined handlers was accepting it. Trying to send the message again...And then send it to the DMQ (if configured)");
            }
        }

        return null;
    }

}
